#include <cstdio>
#include <unistd.h>

/* No IRQ support */

//#define DEBUG

typedef unsigned char BYTE;
typedef unsigned short WORD;
typedef signed char SBYTE;

BYTE m[0x10000]={0};

BYTE A=0,X=0,Y=0,S=255;
WORD PC, WrAddr;
int N,V,/*R,*/B,D,I,Z,C;

BYTE WrB(BYTE b)
{
    m[WrAddr] = b; /* write to WrAddr */
    
    if(WrAddr >= 0x8000
    && WrAddr < 0x8000+40*25)
    {
        unsigned y = (WrAddr-0x8000)/40;
        unsigned x = (WrAddr-0x8000)%40;
        
//#ifndef DEBUG
        printf("\33[%u;%uH", y+1,x+1);
        putchar(b);
        fflush(stdout);
//#endif
    }
#ifdef DEBUG
    fprintf(stderr, "write %02X to %04X\n", b,WrAddr);
#endif
    return b;
}
BYTE RdB(WORD w)
{
    BYTE r = m[w];
   // printf("read %04X -> %02X\n", w,r);
    return r;
}

BYTE P(){return Z|I|C|D|V|N|B|32;} /* R=always 32*/
void F(BYTE r) { Z=r?0:2;N=r&128; }

BYTE MR_ASL(BYTE v) { C=v>127; F(v <<= 1); return v; }
BYTE MR_LSR(BYTE v) { C=v&1;   F(v >>= 1); return v; }
BYTE MR_ROL(BYTE v) { int c=C;return     c  | MR_ASL(v); }
BYTE MR_ROR(BYTE v) { int c=C;return (c<<7) | MR_LSR(v); }

WORD MC_W(BYTE offs)
{
    WORD J = RdB(PC) + 256*RdB(PC+1); PC += 2;
    return J+offs;
}
WORD MC_I(BYTE a,BYTE b)
{
    WORD J = RdB(PC++)+a;
    WORD K=PC; PC=J; J=MC_W(b); PC=K;
    return J;
}
BYTE M_POP() { return RdB(++S | 0x100); }
WORD M_POPW() { WORD J=M_POP();return J+M_POP()*256; }

void M_PUSH(BYTE r)  { WrAddr=0x100|S--; WrB(r); }
void M_PUSHW(WORD w) { M_PUSH(w>>8); M_PUSH(w);  }

void M_CMP(const BYTE r2, const BYTE r1)
{
    register WORD J = r1 - r2; F(J);
    C = J<256;
}
void M_PLP()
{
    int i = M_POP(); F(i);/* sets N */
    C=(i&1);
    Z=(i&2);
    I=(i&4);
    D=(i&8);
    B=(i&16);
    V=(i&64);
}

void emu()
{
    BYTE op = RdB(PC++);
#ifdef DEBUG
    fprintf(stderr, "PC=%04X op=%02X A=%02X X=%02X Y=%02X S=%02X %c%c%c%c%c%c%c\n", 
        PC-1, op, A,X,Y,S,
        N?'N':'-',
        V?'V':'-',
        B?'B':'-',
        D?'D':'-',
        I?'I':'-',
        Z?'Z':'-',
        C?'C':'-');
#endif
    int effa=0;
    switch(
  " x   BB  B   WW "   " y   BB  Y   XX " /* 00,10 */
  "Wx  BBB  B  WWW "   " y   BB  Y   XX " /* 20,30 */
  " x   BB  B  WWW "   " y   BB  Y   XX " /* 40,50 */
  " x   BB  B  WWW "   " y   BB  Y   XX " /* 60,70 */
  " x  BBB     WWW "   " y  BBB  Y   X  " /* 80,90 */
  "BxB BBB  B  WWW "   " y  BBB  Y  XXY " /* A0,B0 */
  "Bx  BBB  B  WWW "   " y   BB  Y   XX " /* C0,D0 */
  "Bx  BBB  B  WWW "   " y   BB  Y   XX " /* E0,F0 */
        [op])
    {
        case 'B': effa=RdB(PC++); break;
        case 'W': effa=MC_W(0); break;
        case 'X': effa=MC_W(X); break;
        case 'Y': effa=MC_W(Y); break;
        case 'x': effa=MC_I(X,0); break;
        case 'y': effa=MC_I(0,Y); break;
    }
    
    WrAddr=effa;
    
    switch(
  " R   RR      RR "   " R   XX  R   RR " /* 00,10 */
  " R  RRR     RRR "   " R   XX  R   RR " /* 20,30 */
  " R   RR      RR "   " R   XX  R   RR " /* 40,50 */
  " R   RR      RR "   " R   XX  R   RR " /* 60,70 */
  "                "   "                " /* 80,90 */
  " R  RRR     RRR "   " R  XX   R  RRR " /* A0,B0 */
  " R  RRR     RRR "   " R   XX  R   RR " /* C0,D0 */
  " R  RRR     RRR "   " R   XX  R   RR " /* E0,F0 */

        [op])
    {
        case 'X': effa=RdB(effa+X); break;
        case 'R': effa=RdB(effa); break;
    }
#ifdef DEBUG__NO
    fprintf(stderr, " - effa=%04X\n", effa);
#endif
    switch(
  "cK   KR nKQ  KR "   "aK   KR bK   KR " /* 00,10 */
  "eL  sLT oLS sLT "   "aL   LT bL   LT " /* 20,30 */
  "pM   MV qMU fMV "   "aM   MV bM   MV " /* 40,50 */
  "di   iXiriW g X "   "ai   iXibi    X " /* 60,70 */
  " m  tmu E C tmu "   "am  vwx AmI  m  " /* 80,90 */
  "OPN OPN BPD OPN "   "aP  OPy bPJ OPN " /* A0,B0 */
  "kh  khY GhF khY "   "ah   hY bh   hY " /* C0,D0 */
  "lj  ljZ Hj  ljZ "   "aj   jZ bj   jZ " /* E0,F0 */
        [op])
    {
        case 'A': F(A=Y);break;         /* TYA */
        case 'B': F(Y=A);break;         /* TAY */
        case 'C': F(A=X);break;         /* TXA */
        case 'D': F(X=A);break;         /* TAX */
        case 'E': F(--Y);break;         /* DEY */
        case 'F': F(--X);break;         /* DEX */
        case 'G': F(++Y);break;         /* INY */
        case 'H': F(++X);break;         /* INX */
        case 'I': S=X;break;              /* TXS */
        case 'J': X=S;break;              /* TSX */
        case 'K': F(A|=effa);break;   /* ORA ($ss),y INDIRINDEX */
        case 'L': F(A&=effa);break;   /* AND ($ss),y INDIRINDEX */
        case 'M': F(A^=effa);break;   /* EOR ($ss),y INDIRINDEX */
        case 'N': F(X=effa);break;   /* LDX #$ss IMM */
        case 'O': F(Y=effa);break;   /* LDY #$ss IMM */
        case 'P': F(A=effa);break;   /* LDA #$ss IMM */
        case 'Q': A=MR_ASL(A);break;         /* ASL a ACC */
        case 'S': A=MR_ROL(A);break;         /* ROL a ACC */
        case 'U': A=MR_LSR(A);break;         /* LSR a ACC */
        case 'W': A=MR_ROR(A);break;         /* ROR a ACC */
        case 'R': WrB(MR_ASL(effa));break;     /* ASL $ssss,x ABS,x */
        case 'T': WrB(MR_ROL(effa));break;     /* ROL $ssss,x ABS,x */
        case 'V': WrB(MR_LSR(effa));break;     /* LSR $ssss,x ABS,x */
        case 'X': WrB(MR_ROR(effa));break;     /* ROR $ssss,x ABS,x */
        case 'Y': F(WrB(effa-1));break;     /* DEC $ssss,x ABS,x */
        case 'Z': F(WrB(effa+1));break;     /* INC $ssss,x ABS,x */
        case 'h': M_CMP(effa,A);break; /* CMP ($ss),y INDIRINDEX */
        case 'k': M_CMP(effa,Y);break; /* CPY #$ss IMM */
        case 'l': M_CMP(effa,X);break; /* CPX #$ss IMM */
        case 'a':
        {
            int*f[4]={&N,&V,&C,&Z};
            if(!*f[op/64] == !(op&32)) PC+=1+(SBYTE)RdB(PC);else++PC;
            break;
        }
        case 'b':
        {
            int*f[4]={&C,&I,&V,&D},v[4]={1,4,2,8},o=op/64;
            *f[o] = v[o]*!!(op&32);
            break;
        }
        case 'e': M_PUSHW(PC-1); /*passthru*/  /* JSR $ssss ABS */
        case 'f': PC = effa; break; /* JMP $ssss ABS */
        case 'g': PC = effa; PC = MC_W(0); break; /* JMP ($ssss) ABDINDIR (IGNORES 6502 BUG) */
        case 'i': /* ADC ($ss),y INDIRINDEX */
        {
            WORD J = A + (BYTE)effa + C; F(J);
            V = (~(A^effa) & (A ^ (BYTE)J) & 128)/2;
            C = J>=256;
            A = (BYTE)J;
            break;
        }
        case 'j': /* SBC ($ss),y INDIRINDEX */
        {
            WORD J = A - (BYTE)effa - !C; F(J);
            V = ((A^effa) & (A ^ (BYTE)J) & 128)/2;
            C = J<256;
            A = (BYTE)J;
            break;
        }
        case 's': F(effa);V=effa&64;Z=(effa&A)?0:2;break;   /* BIT $ssss ABS */

        /*BRK*/
        case 'c': M_PUSHW(++PC);M_PUSH(P()|16);
                  I=4;D=0;B=16;
                  PC = 0xFFFE; PC = MC_W(0);
                  break;

        case 'd': PC = M_POPW(); ++PC; break; /* RTS */
        case 'n': M_PUSH(P());break;        /* PHP */
        case 'o': M_PLP(); break; /* PLP */
        case 'p': M_PLP();PC = M_POPW(); break;
        case 'q': M_PUSH(A);break;        /* PHA */
        case 'r': F(A=M_POP());break;   /* PLA */

        case 'm': WrB(A);break;         /* STA ($ss),y INDIRINDEX */
        case 't': WrB(Y);break;         /* STY $ssss ABS */
        case 'u': WrB(X);break;         /* STX $ssss ABS */
        case 'v': WrAddr=effa+X;WrB(Y);break;         /* STY $ss,x ZP,x */
        case 'w': WrAddr=effa+X;WrB(A);break;         /* STA $ss,x ZP,x */
        case 'x': WrAddr=effa+Y;WrB(X);break;         /* STX $ss,y ZP,y */
        
        case 'y': F(X=RdB(effa+Y));break;   /* LDX $ss,y ZP,y */        
        // Bad ops should be silently ignored like NOP.
    }
}

int main(void)
{
#if 1 /* PET test */
    fread(m+0xC000, 1,0x8000,stdin);
    
    m[0xEFD8] = 0xFF;
    
    PC=0xC000;
#else
    fread(m+0x8000-0x80,1,0x8000,stdin);
    
    PC = 0x8000-0x80+0x0A; PC = MC_W(0);
#endif
    while(m[PC] != 0x60 || S != 255) emu(); /* Until RTS */
    
    printf("init done\n");
    
    PC = 0x8000-0x80+0x0C; PC = MC_W(0);
    for(;;)
    {
        emu();
        usleep(1000000/60);
    }
}
